//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File: diag_hash.c
//
//----------
//
// diag_hash--
//	Support for managing a hash table of values associated with alignment
//	diagonals.
//
// Our search algorithms need to track hits on each diagonal to detect
// overlapping hits and nearby multiple hits.  For each diagonal, we store the
// position in the second sequence of the end of most recent hit (pos2), and in
// some cases we also store the start position of a run of overlapping or
// nearly-overlapping hits.
//
// Ideally we would have an entry for each possible diagonal.  But with
// sequence lengths approaching a quarter billion, such an array would require
// too much memory.  Instead, we use a hash array much smaller than the number
// of hits likely to be "active" at any one time.  It is critical that we store
// pos2 for this purpose rather than pos1, because pos2 always increases during
// the search while pos1 hops around the sequence.  By guaranteeing the order
// that we (effectively) scan each diagonal, we limit our exposure to collisions
// in the hash.  We also store the number of the actual diagonal associated with
// each hash location, so that we can detect collisions.
//
// The effect of a collision is that a hit that should be combined with a
// nearby one to its left won't be.  Because the second sequence is being
// scanned from left to right, a hit stays "dangerous" (i.e., it can cover up
// a subsequent hit on a different hash-equivalent diagonal) only until the
// sweep along sequence 2 reaches pos2 for the end of that hit.  If we were
// storing pos1, a hit could stay dangerous for the entire sweep.
//
// The hash funtion used simply uses some number of the least significant bits
// of the diagonal.  For this reason the hash size H (diagHashSize) must be a
// power of 2.  Hash-equivalent diagonals are not random but are separated by H
// intervening diagonals.
//
// Helpful formulae:
//	diag = pos1 - pos2
//	pos1 = diag + pos2
//	pos2 = pos1 - diag
//
// We could reduce memory use by using two (or more) hashes with different
// moduli (relatively prime).  The tradeoff is that we'd be spending more time
// computing the hash values and reading/writing them from/to the hash arrays.
//
//----------

//----------
//
// other files
//
//----------

#include <stdlib.h>				// standard C stuff
#define  true  1
#define  false 0

#define  diag_hash_owner		// (make this the owner of its globals)
#include "diag_hash.h"			// interface to this module

//----------
//
// empty_diag_hash--
//	Create or re-create an empty diagonals hash.
//
//----------
//
// Arguments:
//	(none)
//
// Returns:
//	(nothing)
//
//----------
//
// Notes:
//
// (1)	This routine allocates and reuses memory via global pointers.  The
//		caller should make a call to free_diag_hash() to de-allocate this
//		memory when it is no longer needed.
//
//----------

void empty_diag_hash (void)
	{
	// allocate (or re-use) memory;  note that only diagEnd requires that we
	// fill it with zeros

	if (diagEnd == NULL)
		{
		// allocate new array
		// caveat: sometimes we don't need all of this (e.g. process_for_simple_hit)
		//         .. so allocating all these together is wasteful

		diagEnd    = zalloc_or_die ("empty_diag_hash (diagEnd)",
		                             diagHashSize * sizeof(unspos));
		diagStart  = malloc_or_die ("empty_diag_hash (diagStart)",
		                             diagHashSize * sizeof(unspos));
		diagActual = malloc_or_die ("empty_diag_hash (diagActual)",
		                             diagHashSize * sizeof(sgnpos));
		diagActive = malloc_or_die ("empty_diag_hash (diagActive)",
		                             diagHashSize * sizeof(int));
		numDiagActive = 0;
		}
	else
		{
		// clear the previously-allocated data

		while (numDiagActive > 0)
			diagEnd[diagActive[--numDiagActive]] = 0;
		}
	}

//----------
//
// free_diag_hash--
//	Dispose of the diagonals hash.
//
//----------
//
// Arguments:
//	(none)
//
// Returns:
//	(nothing)
//
//----------

void free_diag_hash
   (void)
	{
	free_if_valid ("free_diag_hash (diagEnd)",    diagEnd);    diagEnd    = NULL;
	free_if_valid ("free_diag_hash (diagStart)",  diagStart);  diagStart  = NULL;
	free_if_valid ("free_diag_hash (diagActual)", diagActual); diagActual = NULL;
	free_if_valid ("free_diag_hash (diagActive)", diagActive); diagActive = NULL;
	}

