//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File: masking.c
//
//----------
//
// masking--
//	Provide dynamic masking for sequences that do not have their repeats
//	masked.
//
// The basic premise is that if anything in the first sequence aligns to a
// repeat in the second sequence, it will align to all copies of the repeat.
// Thus, positions that align frequently are probably repeat elements.
//
// For every position in sequence 1, we keep a count of how many times it has
// been part of an alignment (even if it's part of an indel).  When the count
// reaches the given threshold, we mark that position with an 'x' so that
// subsequent alignments will give it a large scoring penalty.
//
// This is most useful if the caller is comparing sequence 1 to many other
// sequences.  The first comparison(s) will get no advantage-- nothing will be
// masked.  But if enough sequences are compared, the later sequences will
// benefit from the masking.
//
//----------

//----------
//
// other files
//
//----------

#include <stdlib.h>				// standard C stuff
#define  true  1
#define  false 0
#include <string.h>				// standard C string stuff
#include "utilities.h"			// utility stuff
#include "segment.h"			// segment table management stuff
#include "edit_script.h"		// alignment edit script stuff

#define  masking_owner			// (make this the owner of its globals)
#include "masking.h"			// interface to this module

//----------
//
// miscellaneous data
//
//----------

//#define noCallback			// if this is defined, mask_interval doesn't
								// bother to call the callback routine for
								// masked intervals;  this is to test whether
								// there is any speed advantage to removing
								// masked seeds from the position table;  note
								// that there are other good reasons to remove
								// masked seeds, regardless of speed

//----------
//
// prototypes for private functions
//
//----------

static unspos mask_interval (u8* fwd, u8* rev, unspos beg, unspos end,
                             census* cen,
                             void(*func) (unspos beg, unspos end, void* info),
                             void* info);

//----------
//
// new_census--
//	Create a new cencus of a given length.
//
//----------
//
// Arguments:
//	new_census	len:		The length of the sequence to be censused.
//	u8			maskThresh:	The count threshold.
//
// Returns:
//	Pointer to the new census (failures are fatal);  caller must eventually
//	dispose of this memory with a call to free().
//
//----------

census* new_census
   (unspos	len,
	u8		maskThresh)
	{
	census*	cen;
	size_t	bytesNeeded;

	if (len == 0) len = 1;
	bytesNeeded = sizeof(census) + ((len-1)*sizeof(u8));
	cen = (census*) zalloc_or_die ("new_census", bytesNeeded);

	cen->len        = len;
	cen->maskThresh = maskThresh;

	return cen;
	}

//----------
//
// census_mask_segments--
//	Given a list of ungapped alignment segments, increment the count of all
//	aligned locations and for any that meet or exceed the threshold, mask them
//	from the sequence.
//
//----------
//
// Arguments:
//	segtable*	st:			The segments.
//	u8*			fwd, rev:	The sequence and its reverse.  rev may be NULL.
//	census*		cen:		The census to update.
//	void (*func):			A callback function to report masked intervals to.
//	        (unspos beg,	.. beg,end is the interval that is about to be
//	         unspos end,	.. masked (origin 1, inclusive).  info is passed
//	         void* info)	.. through from our caller.
//	void*		info:		Pass-thru for func.
//
// Returns:
//	A count of the number of bases masked.  This does not include any bases that
//	were previously masked.
//
//----------

unspos census_mask_segments
   (segtable*	st,
	u8*			fwd,
	u8*			rev,
	census*		cen,
	void		(*func)(unspos beg, unspos end, void* info),
	void*		info)
	{
	segment*	seg;
	u32			ix;
	unspos		beg, end, i;
	unspos		count = 0;

	if (cen == NULL)
		return 0;

	for (ix=0,seg=st->seg ; ix<st->len ; ix++,seg++)
		{
		if ((seg->length > cen->len) || (seg->pos1 > cen->len - seg->length))
			suicide ("census_mask_segments, internal error");
		beg = seg->pos1;
		end = beg + seg->length;
		for (i=beg ; i<end ; i++)
			{ if (cen->count[i] < 255) cen->count[i]++; }
		if (cen->maskThresh > 0)
			count += mask_interval (fwd, rev, beg+1, end, cen, func, info);
		}

	masking_add_stat (maskedBases, count);
	return count;
	}

//----------
//
// census_mask_aligns--
//	Given a list of alignments, increment the count of all aligned locations
//	and for any that meet or exceed the threshold, mask them from the sequence.
//
// Note that the entire aligned interval is counted/masked, even indels.
//
//----------
//
// Arguments:
//	alignel*	alignList:	The alignments.
//	u8*			fwd, rev:	The sequence and its reverse.  rev may be NULL.
//	census*		cen:		The census to update.
//	void (*func):			A callback function to report masked intervals to.
//	        (unspos beg,	.. beg,end is the interval that is about to be
//	         unspos end,	.. masked (origin 1, inclusive).  info is passed
//	         void* info)	.. through from our caller.
//	void*		info:		Pass-thru for func.
//
// Returns:
//	A count of the number of bases masked.  This does not include any bases that
//	were previously masked.
//
//----------

unspos census_mask_aligns
   (alignel*	alignList,
	u8*			fwd,
	u8*			rev,
	census*		cen,
	void		(*func)(unspos beg, unspos end, void* info),
	void*		info)
	{
	alignel*	a;
	unspos			beg, end, i;
	unspos		count = 0;

	if (cen == NULL)
		return 0;

	for (a=alignList ; a!=NULL ; a=a->next)
		{
		if (a->beg1 < 1)
			suicide ("census_mask_aligns, internal error");
		if (a->end1 > cen->len)
			suicide ("census_mask_aligns, internal error");
		beg = a->beg1 - 1;
		end = a->end1;
		for (i=beg ; i<end ; i++)
			{ if (cen->count[i] < 255) cen->count[i]++; }
		if (cen->maskThresh > 0)
			count += mask_interval (fwd, rev, beg+1, end, cen, func, info);
		}

	masking_add_stat (maskedBases, count);
	return count;
	}

//----------
//
// mask_interval--
//	Given an interval, mask any position that meets or exceeds the count
//	threshold.
//
//----------
//
// Arguments:
//	u8*			fwd, rev:	The sequence and its reverse.  rev may be NULL.
//	unspos		beg, end:	The interval to check.  Origin-1, inclusive.
//	census*		cen:		The census describing which bases to mask.
//	void (*func):			A callback function to report masked intervals to.
//	        (unspos beg,	.. beg,end is the interval that is about to be
//	         unspos end,	.. masked, see note 1 (origin 1, inclusive).  info
//	         void* info)	.. is passed through from our caller.
//	void*		info:		Pass-thru for func.
//
// Returns:
//	A count of the number of bases masked.  This does not include any bases that
//	were previously masked.
//
//----------
//
// Notes:
//	(1)	The intervals passed to the callback function are not the same as the
//		interval passed to this routine.  Only sub-intervals that reach the
//		masking threshold for the first time are passed to the callback.
//
//----------

#ifndef noCallback		// normal version, callback used

static unspos mask_interval
   (u8*		fwd,
	u8*		rev,
	unspos	beg,
	unspos	end,
	census*	cen,
	void	(*func)(unspos beg, unspos end, void* info),
	void*	info)
	{
	static const u32 noRun = (u32) -1;
	unspos	revLen, runBeg, i, j;
	unspos	count = 0;

	if (cen == NULL)
		return 0;

	if ((beg < 1) || (end > cen->len))
		suicide ("mask_interval, internal error");

	revLen = cen->len - 1;

	runBeg = noRun;
	for (i=beg-1 ; i<end ; i++)
		{
		if ((cen->maskThresh > 0)
		 && (cen->count[i] >= cen->maskThresh)
		 && (dna_isupper(fwd[i])))
			{
			if (runBeg == noRun) runBeg = i;
			}
		else if (runBeg != noRun)
			{
			func (runBeg+1, i, info);
			for (j=runBeg ; j<i ; j++)
				{
				fwd[j] = 'x';  count++;
				if (rev != NULL) rev[revLen-j] = 'x';
				}
			runBeg = noRun;
			}
		}

	if (runBeg != noRun)
		{
		func (runBeg+1, end, info);
		for (j=runBeg ; j<end ; j++)
			{
			fwd[j] = 'x';  count++;
			if (rev != NULL) rev[revLen-j] = 'x';
			}
		}

	return count;
	}

#endif // not noCallback


#ifdef noCallback		// test version, callback not used

static int mask_interval
   (u8*		fwd,
	u8*		rev,
	unspos	beg,
	unspos	end,
	census*	cen,
	void	(*func)(unspos beg, unspos end, void* info),
	void*	info)
	{
	unspos	revLen, i;
	unspos	count = 0;

	if (cen == NULL)
		return 0;

	if ((beg < 1) || (end > cen->len))
		suicide ("mask_interval, internal error");

	revLen = cen->len - 1;
	for (i=beg-1 ; i<end ; i++)
		{
		if ((cen->maskThresh > 0)
		 && (cen->count[i] >= cen->maskThresh))
			{
			if (dna_isupper(fwd[i])) { fwd[i] = 'x';  count++; }
			if (rev != NULL)         rev[revLen-i] = 'x';
			}
		}

	return count;
	}
	
#endif // noCallback

//----------
//
// report_census_intervals--
//	Report runs in a census of positions that meet or exceed the threshold.
//
//----------
//
// Arguments:
//	census*		cen:		The census to report.
//	void (*func):			The reporting function.  beg,end is the interval
//	        (unspos beg,	.. (origin 1, inclusive).  info is passed through
//	         unspos end,	.. from our caller.
//	         void* info)
//	void*		info:		Pass-thru for func.
//
// Returns:
//	The number of intervals reported.
//
//----------

unspos report_census_intervals
   (census*	cen,
	void	(*func)(unspos beg, unspos end, void* info),
	void*	info)
	{
	static const u32 noRun = (u32) -1;
	unspos	runBeg, i;
	unspos	count = 0;

	if (cen == NULL)
		return 0;

	runBeg = noRun;
	for (i=0 ; i<cen->len ; i++)
		{
		if (cen->count[i] >= cen->maskThresh)
			{
			if (runBeg == noRun) runBeg = i;
			}
		else if (runBeg != noRun)
			{
			func (runBeg+1, i, info);
			count++;
			runBeg = noRun;
			}
		}

	if (runBeg != noRun)
		{
		func (runBeg+1, cen->len, info);
		count++;
		}

	return count;
	}

//----------
//
// print_census--
//	Print all locations in a census that meet or exceed the threshold.
//
//----------
//
// Arguments:
//	FILE*	f:		The file to print to.
//	census*	cen:	The census to print.
//
// Returns:
//	(nothing)
//
//----------

void print_census
   (FILE*	f,
	census*	cen)
	{
	unspos	i;

	if (cen == NULL) return;

	for (i=0 ; i<cen->len ; i++)
		{
		if (cen->count[i] >= cen->maskThresh)
			fprintf(f, unsposFmt " %u\n", i+1, cen->count[i]);
		}
	}

//----------
//
// masking_zero_stats--
//	Clear the statistics for this module.
//
//----------
//
// Arguments:
//	(none)
//
// Returns:
//	(nothing)
//
//----------

void masking_zero_stats
   (void)
	{
#ifdef collect_stats

	// set 'em en masse to zero

	memset (&maskingStats, 0, sizeof(maskingStats));

	// set any values that might be floating point to zero (fp bit pattern for
	// zero may not be all-bits-zero)

	// (none to set, yet)

#endif // collect_stats
	}

//----------
//
// masking_show_stats--
//	Show the statistics that have been collected for this module.
//
//----------
//
// Arguments:
//	FILE*		f:	The file to print the stats to.
//
// Returns:
//	(nothing)
//
//----------

void masking_show_stats
   (arg_dont_complain(FILE* f))
	{
#ifdef collect_stats
	if (f == NULL) return;
	fprintf (f, "      masked bases: %s\n", commatize(maskingStats.maskedBases));
	fprintf (f, "-------------------\n");
#endif // collect_stats
	}

void masking_generic_stats
   (arg_dont_complain(FILE* f),
    arg_dont_complain(void (*func) (FILE*, const char*, ...)))
	{
#ifdef collect_stats
	if (f == NULL) return;
	(*func) (f, "masked_bases=%d\n", maskingStats.maskedBases);
#endif // collect_stats
	}

