//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File: utilities.h
//
//----------

#ifndef utilities_H				// (prevent multiple inclusion)
#define utilities_H

#include <stdio.h>				// standard C i/o stuff

//----------
//
// data structures and types
//
//----------

// sized data types;  these generally come from stdint.h but on some older
// platforms it may not exist, in which case override_stdint can be enabled
// and the proper types set up here

#ifdef override_stdint

typedef signed   char      int8;
typedef signed   char      s8;
typedef unsigned char      u8;
typedef          long int  int32;
typedef          long int  s32;
typedef unsigned long int  u32;
typedef          long long int64;
typedef          long long s64;
typedef unsigned long long u64;

#else

#include <stdint.h>
typedef int8_t   int8;
typedef int8_t   s8;
typedef uint8_t  u8;
typedef int32_t  int32;
typedef int32_t  s32;
typedef uint32_t u32;
typedef int64_t  int64;
typedef int64_t  s64;
typedef uint64_t u64;

#endif // override_stdint

// short strings

typedef struct char3 { char s[4]; } char3;

// macro to round data structure sizes to the next larger multiple of 8, 16,
// or m

#define round_up_2(b)   ((((unsigned) (b))+1)&(~1))
#define round_up_4(b)   ((((unsigned) (b))+3)&(~3))
#define round_up_8(b)   ((((unsigned) (b))+7)&(~7))
#define round_up_16(b)  ((((unsigned) (b))+15)&(~15))
#define round_up_32(b)  ((((unsigned) (b))+31)&(~31))
#define round_up_64(b)  ((((unsigned) (b))+63)&(~63))
#define round_up_128(b) ((((unsigned) (b))+127)&(~127))
#define round_up_256(b) ((((unsigned) (b))+255)&(~255))
#define round_up_512(b) ((((unsigned) (b))+511)&(~511))
#define round_up_1K(b)  ((((unsigned) (b))+1023)&(~1023))
#define round_up_2K(b)  ((((unsigned) (b))+2047)&(~2047))
#define round_up_4K(b)  ((((unsigned) (b))+4095)&(~4095))
#define round_up_8K(b)  ((((unsigned) (b))+8191)&(~8191))
#define round_up_16K(b) ((((unsigned) (b))+16383)&(~16383))
#define round_up(b,m)   (((((unsigned) (b))+((m)-1))/(m))*(m))

// macro to count the number of entries in a staticly declared array

#define entriesof(array) (sizeof(array)/sizeof((array)[0]))

// silly type check defeaters

#define ustrlen(s)     (strlen((char*)(s)))
#define ustrcmp(s1,s2) (strcmp((char*)(s1),(char*)(s2)))
#define ustrcpy(s1,s2) (strcpy((char*)(s1),(char*)(s2)))
#define ustrchr(s,c)   (strchr((char*)(s),(char)(c)))

#define strleni(s)     ((int)(strlen(s)))

// macro to convince gnu c compiler not to complain about unusued function
// arguments

#ifdef __GNUC__
#define arg_dont_complain(arg) arg __attribute__ ((unused))
#else
#define arg_dont_complain(arg) arg
#endif // __GNUC__

//----------
//
// memory allocation routines in utilities.c
//
// These routines wrap the memory allcoation routines in the standard library,
// and permit us to use some post-processing tools to inspect memory usage.
// They are affected by two compile-time #defines:
//
//		trackMemoryUsage:	If defined, the routines will write a detailed
//							memory de/re/allocation history to stderr, which
//							can be processed by memory_sniffer (part of the
//							lastz tools).
//
//		noMemoryWrappers:	If defined, calls to the memory wrappers are
//							replaced by calls to the standard library routines.
//							This is useful in conjunction with Valgrind's
//							heap profiler, Massif.
//
//----------

#ifdef noMemoryWrappers
#define malloc_or_die(id,size)    malloc (size)
#define zalloc_or_die(id,size)    calloc (1,size)
#define realloc_or_die(id,p,size) realloc (p,size)
#define free_if_valid(id,p)       free (p)
#endif // noMemoryWrappers

#ifndef noMemoryWrappers
void* malloc_or_die  (char* id, size_t size);
void* zalloc_or_die  (char* id, size_t size);
void* realloc_or_die (char* id, void* _p, size_t size);
void  free_if_valid  (char* id, void* p);
#endif // not noMemoryWrappers


#ifdef trackMemoryUsage
#define memory_checkpoint(fmt)       fprintf(stderr,fmt)
#define memory_checkpoint_1(fmt,i)   fprintf(stderr,fmt,i)
#define memory_checkpoint_2(fmt,i,s) fprintf(stderr,fmt,i,s)
#endif // trackMemoryUsage

#ifndef trackMemoryUsage
#define memory_checkpoint(fmt)       ;
#define memory_checkpoint_1(fmt,i)   ;
#define memory_checkpoint_2(fmt,i,s) ;
#endif // not trackMemoryUsage

//----------
//
// prototypes for routines in utilities.c
//
//----------

FILE*  fopen_or_die         (const char* name, const char* mode);
int    fclose_if_valid      (FILE* f);
int    getc_or_die          (FILE* f);
int    print_prefix         (FILE* f, const char* s, int n);
char*  copy_string          (const char* s);
char*  copy_prefix          (const char* s, int n);
void   append_char          (char** s, u32* size, u32* len, char ch);
void   append_u8            (u8**   s, u32* size, u32* len, u8   ch);
int    strcmp_prefix        (const char* str1, const char* str2);
int    strcmp_suffix        (const char* str1, const char* str2);
int    is_blank_string      (const char* s);
int    string_to_int        (const char* s);
int    string_to_int2       (const char* s, int byThousands);
int    hex_string_to_int    (const char* s);
double string_to_double     (const char* s);
double pct_string_to_double (const char* s);
char3  prob_to_string       (double p);
int    string_replace       (char* s, int len, char* sub, char* rep);
char*  trim_string          (char* s);
char*  skip_whitespace      (char* s);
char*  skip_darkspace       (char* s);
char*  skip_til             (char* s, char* chars);
char*  skip_while           (char* s, char* chars);
char*  reverse_of_string    (char* s, int len);
char*  commatize            (const int64 v);
char*  unitize              (const int64 v);
int    bit_count            (u32 bits);
int    bit_count_64         (u64 bits);
int    bit_count_16         (u32 bits);
int    bit_count_8          (u8 bits);
void   suicide              (const char* message);
void   suicidef             (const char* format, ...);

#endif // utilities_H
